/*
 * @Author: your name
 * @Date: 2021-11-16 09:18:58
 * @LastEditTime: 2021-12-16 11:15:44
 * @LastEditors: Please set LastEditors
 * @Description: In User Settings Edit
 * @FilePath: /roadmodel/src/CtApRoadModel/source/road/obstacle.h
 */
#pragma once
#include <stdint.h>

#include <list>

#include "Eigen/Dense"
#include "boundary.h"
#include "corridor.h"
#include "debouncer.h"
#include "delay.h"
#include "enum_list.h"
#include "extend_kalman_filter_object.h"
#include "flip_flop.h"
#include "global_parameters.h"
#include "lane.h"
#include "lane_change_intention_detector.h"
#include "lookup_table.h"
#include "observer.h"
#include "rm_config.h"
#include "sl_point.h"
#include "vec2d.h"
#include "vehicle_state_provider.h"
using rm::common::math::Vec2d;

namespace rm {
struct InCorridorDDStruct {
  InCorridorDDStruct(Delay<bool> in_corridor_delay_,
                     Delay<bool> oncoming_delay_)
      : left_debouncer(Debouncer()), right_debouncer(Debouncer()),
        in_corridor_delay(in_corridor_delay_),
        oncoming_delay(oncoming_delay_){};
  InCorridorDDStruct(){};
  rm::Debouncer left_debouncer;
  rm::Debouncer right_debouncer;
  rm::Delay<bool> in_corridor_delay;
  rm::Delay<bool> oncoming_delay;
};

struct ObjHighlightDDStruct {
  rm::Debouncer lat_dis_debouncer;
  rm::Debouncer long_dis_debouncer;
  bool excuted_flag;
};

class Obstacle {
public:
  Obstacle() {
    // reassign value to make dynamic config effect
    in_corridor_long_dist_inner_table_ = OneDimLookupTable(
        GetRMConfig().in_corridor_long_dist_inner_table_x,
        GetRMConfig().in_corridor_long_dist_inner_table_y); // factor
    in_corridor_long_dist_outer_table_ = OneDimLookupTable(
        GetRMConfig().in_corridor_long_dist_outer_table_x,
        GetRMConfig().in_corridor_long_dist_outer_table_y); // factor
    in_corridor_inner_width_table_ = OneDimLookupTable(
        GetRMConfig().in_corridor_inner_width_table_x,
        GetRMConfig().in_corridor_inner_width_table_y); // base
    in_corridor_outer_width_table_ =
        OneDimLookupTable(GetRMConfig().in_corridor_outer_width_table_x,
                          GetRMConfig().in_corridor_outer_width_table_y);
    in_corridor_ego_position_left_table_ =
        OneDimLookupTable(GetRMConfig().in_corridor_ego_position_left_table_x,
                          GetRMConfig().in_corridor_ego_position_left_table_y);
    in_corridor_ego_position_right_table_ =
        OneDimLookupTable(GetRMConfig().in_corridor_ego_position_right_table_x,
                          GetRMConfig().in_corridor_ego_position_right_table_y);
    in_corridor_ego_position_table_ =
        OneDimLookupTable(GetRMConfig().in_corridor_ego_position_table_x,
                          GetRMConfig().in_corridor_ego_position_table_y);
    width_factor_dl_car_ =
        OneDimLookupTable(GetRMConfig().width_factor_dl_car_x,
                          GetRMConfig().width_factor_dl_car_y);
    width_factor_dl_trunck_ =
        OneDimLookupTable(GetRMConfig().width_factor_dl_trunck_x,
                          GetRMConfig().width_factor_dl_trunck_y);
    in_corridor_upper_long_dis_ego_velocity_table_ = OneDimLookupTable(
        GetRMConfig().in_corridor_upper_long_dis_ego_velocity_table_x,
        GetRMConfig().in_corridor_upper_long_dis_ego_velocity_table_y);
    in_corridor_lower_long_dis_ego_velocity_table_ = OneDimLookupTable(
        GetRMConfig().in_corridor_lower_long_dis_ego_velocity_table_x,
        GetRMConfig().in_corridor_lower_long_dis_ego_velocity_table_y);
    width_factor_s_ = OneDimLookupTable(GetRMConfig().width_factor_s_x,
                                        GetRMConfig().width_factor_s_y);
    width_factor_ego_yaw_rate_table_ =
        OneDimLookupTable(GetRMConfig().width_factor_ego_yaw_rate_table_x,
                          GetRMConfig().width_factor_ego_yaw_rate_table_y);
  }
  void set_id(uint32_t id);
  void set_exist_probability(float exist_probability);
  void set_classification(ObstacleType classification);
  void set_position(Vec2d position);
  void set_center(Vec2d center);
  void set_length(float length);
  void set_width(float width);
  // void set_heading(float heading);
  void set_heading(float orientation);
  void set_yaw_rate(float yaw_rate);
  void set_speed(Vec2d speed);
  void set_acceleration(Vec2d acceleration);
  void set_kappa(float kappa);
  void set_moving_direction(uint8_t movingdirection);
  void set_motion_property(uint8_t motion_property);
  void set_is_front_obstacle(bool is_front_obstacle);
  void set_prv_corridor_id(uint32_t prv_corridor_id);
  void unset_prv_corridor_id();
  void set_is_already_assigned();
  void unset_is_already_assigned();
  void set_reference_point_type(ReferencePointType reference_point_type);
  void Set_existence();
  void Unset_existence();
  void reset_trplontype();

  uint32_t id() const;
  float exist_probability() const;
  ObstacleType classification() const;
  Vec2d position() const;
  Vec2d center() const;
  float length() const;
  float width() const;
  float heading() const;
  float yaw_rate() const;
  Vec2d speed() const;
  Vec2d acceleration() const;
  ObstacleMotionProperty motion_property() const;
  ObstacleMovingDirection moving_direction() const;
  float kappa() const;
  bool existence() const;
  const Corridor &virtual_corridor() const;
  const std::list<Vec2d> &historical_traj() const;
  bool lane_change_intention() const;
  friend std::ostream &operator<<(std::ostream &os, const Obstacle &obstacle);
  void Update_pos(double angle, double dx); // update the history pos point
                                            // according to ego vehicle motion
  void Record_pos_center(Vec2d newpos, Vec2d newcenter);
  void Set_kappa();
  void Init_EKF(const std::vector<double> &measurement);
  void Run_EKF(const std::vector<double> &measurement);
  void Run_lane_change_intention_detector(const ObjectLocation &obj_loc);
  // bool In_other_corridor_check(const float lane_width, const SLPoint
  // &object);
  bool In_other_corridor_check(const Lane &lane, const Corridor *corridor,
                               const SLPoint &object);
  bool In_ego_corridor_check(const Corridor *const corridor,
                             const rm::Lane &lane, const SLPoint &object);
  bool In_ego_corridor_check(const Corridor *const corridor,
                             const SLPoint &object, double ego_offset,
                             const rm::Lane &lane, bool lane_changed_delay);
  void Run_in_corridor_long_distance_debouncer(double ego_velocity_kph);
  bool Run_rear_delay(const bool rear_flag);
  void Run_moving_delay();
  void Run_is_riding_marker_delay(const ObjectLocation &obj_loc,
                                  float lane_width);
  void OvertakingJudge(const ObjectLocation &obj_loc, float ref_velocity);
  void Run_is_parallel_delay(const ObjectLocation &obj_loc);
  bool Generate_virtual_corridor();
  bool Generate_virtual_corridor(
      const std::pair<Boundary, Boundary> *lane_info_boundary);
  void Check_virtual_corridor_generation_plausibility();
  bool operator==(const uint32_t id) const;
  void Bind_with_lane(uint8_t lane_id);
  void DeleteandReset_incorridor_map();
  void Update_incorridor_map_key(uint8_t ego_lane_id_last_cycle,
                                 uint8_t ego_lane_id);
  void set_trplontype(TargetType target_type);
  TargetType trplontype() const;
  TargetType prv_trplontype() const;
  bool delayed_motion() const;
  bool is_in_long_distance_range() const;
  bool is_front_obstacle() const;
  bool is_riding_on_lane_marker() const;
  bool is_overtaking() const;
  bool is_parallel_obstacle() const;
  float existence_duration() const;
  void Update_existence_duration();
  bool is_already_assigned() const;
  uint32_t prv_corridor_id() const;
  void Add_duration_of_intermidiate_status();
  void Reset_duration_of_intermidiate_status();
  void Reset_lane_change_intention_detector();
  float duration_of_intermidiate_status();
  ReferencePointType reference_point_type() const;
  bool Run_oncoming_delay(const ObjectLocation &obj_loc, const uint8_t lane_id);
  bool is_front_adjacent_target() const;
  void Reset_was_in_ego_corridor_assured_flfp();
  void Reset_vulnerable_object_critical_ego_corridor_debouncer_pair_pre();
  void Reset_vulnerable_object_critical_ego_corridor_debouncer_pair_post();
  void set_is_vulnerable_critical_obstacle(bool flag);
  bool is_vulnerable_critical_obstacle() const;
  static void Compute_prediction_factor_ego_yaw_rate();

private:
  uint32_t id_;
  float exist_probability_;     //目标存在的概率
  ObstacleType classification_; //目标类别
  Vec2d position_;              // the reference point coordinate
  Vec2d center_;                // the center point coordinate
  float length_;                //目标长度
  float width_;                 //目标宽度
  float heading_;      //目标的航向角 自车坐标系 逆时针为正
  float yaw_rate_;     // object yaw rate
  Vec2d speed_;        //目标相对自车坐标系的速度(vx vy)
  Vec2d acceleration_; //目标相对自车坐标系的加速度(ax ay)
  float kappa_{0.0};        //曲率
  float existence_duration_ = 0.0; // duration since first apperance
  ObstacleMovingDirection
      moving_direction_; // driving direction if motion_property is moving,
                         // otherwise indicating the historical moving
                         // direction
  ObstacleMotionProperty motion_property_; //
  std::list<Vec2d> historical_traj_;       // historical trajectory
  Corridor virtual_corridor_;              // virtual corridor of the object
  bool existence_flag_ = false; // flag indicating the existence of object
  rm::common::math::ExtendKalmanFilterObject ekfilter_;
  TargetType trplontype_ =
      TargetType::OTHER; // target type used for longitudinal plan
  TargetType prv_trplontype_ = TargetType::OTHER;
  bool trplontype_changed_ = false;
  bool delayed_motion_;
  bool is_front_obstacle_;
  bool is_riding_on_lane_marker_;
  bool is_overtaking_{false};
  bool is_parallel_obstacle_;
  bool is_in_long_distance_range_;
  std::array<double, 6> filtered_state_;
  uint32_t prv_corridor_id_;
  bool is_already_assgined_ = false;
  float duration_of_intermidiate_status_;

  // std::map<uint32_t, std::pair<std::pair<std::array<Debouncer, 2> ,
  // Delay<bool>>, bool>> incorridor_debouncer_delay_map_; // the idx of the map
  // is the id of the corridor
  std::map<uint8_t, std::pair<InCorridorDDStruct, bool>>
      incorridor_debouncer_delay_map_;
  Debouncer incorridor_debouncer_; // debouncer for in-corridor check
  Debouncer max_longdistance_debouncer_ = {
      50, 52.5}; // debouncer for long. distance check
  Debouncer min_longdistance_debouncer_ = {
      0.0, 1.0}; // debouncer for long. distance check
  Debouncer in_corridor_long_debouncer_;
  Delay<bool> incorridor_debouncer_delay_ = {0.08, 0.0, false};
  Delay<bool> longdistance_debouncer_delay_ = {0.1, 0.0, false};
  Delay<bool> delay_rear_ = {0.16, 0.0, false};
  Delay<bool> moving_delay_ = {0.8, 0.12, false};
  Delay<bool> parallel_delay_ = {1.0, 0.5, false};
  Delay<bool> riding_delay_ = {2.0, 0.5, false};
  Delay<bool> overtaking_delay_ = {0.4, 0.2, false};
  static OneDimLookupTable in_corridor_long_dist_inner_table_;
  static OneDimLookupTable in_corridor_long_dist_outer_table_;
  static OneDimLookupTable in_corridor_inner_width_table_;
  static OneDimLookupTable in_corridor_outer_width_table_;
  static OneDimLookupTable in_corridor_ego_position_left_table_;
  static OneDimLookupTable in_corridor_ego_position_right_table_;
  static OneDimLookupTable in_corridor_ego_position_table_;
  static OneDimLookupTable in_corridor_lower_long_dis_ego_velocity_table_;
  static OneDimLookupTable in_corridor_upper_long_dis_ego_velocity_table_;
  static OneDimLookupTable width_factor_dl_car_;
  static OneDimLookupTable width_factor_dl_trunck_;
  static OneDimLookupTable width_factor_s_;
  static OneDimLookupTable width_factor_ego_yaw_rate_table_;
  static double width_factor_ego_yaw_rate_;
  static const int max_traj_points_;
  rm::ReferencePointType reference_point_type_;
  bool lane_change_intention_ = false; // lane change intention w.r.t ego lane
  LaneChangeIntentionDetector lane_change_intention_detector_;
  bool is_front_adjacent_target_;
  bool virtual_corridor_generation_plausibility_;
  bool was_in_ego_corridor_assured_;
  common::FlipFlop was_in_ego_corridor_assured_flfp_;
  ObjHighlightDDStruct vulnerable_object_critical_ego_corridor_debouncer_pair;
  bool is_vulnerable_critical_obstacle_;

  std::shared_ptr<Observer> observer_;
};

} // namespace rm